#include "projutil.h"
#include <vsapi.h>
#include "std.h"
#include "str.h"
#include "arrays.h"
#include "flatfile.h"
#include "icons.h"
#include "res.h"

static int s_iMode = 1;
static Rg<FileInfo>* s_rgprgFiles[3] = { 0 };
static Rg<FileInfo*>* s_prgMatches = 0;

static bool s_fSortByPath = false;
static bool s_fAlwaysSubstringMatch = false;

/*
 * FileInfo.
 */

FileInfo::FileInfo()
	: m_iImage( II_NONE )
	, m_fEnsuredIcon( false )
{
}

bool
FileInfo::EnsureSysIcon( HWND hwndNotify )
{
	if( !m_fEnsuredIcon )
	{
		Assert( !m_pAsyncShellIcon );

		m_fEnsuredIcon = true;
		m_pAsyncShellIcon = new AsyncShellIcon();

		if( m_pAsyncShellIcon )
		{
			Str s;
			s.Set( m_sPath );
			s.Append( m_sFile );
			m_pAsyncShellIcon->Enqueue( hwndNotify, s.Text(), SHGFI_SMALLICON );
		}
	}

	if( m_pAsyncShellIcon && m_pAsyncShellIcon->IsComplete() )
	{
		m_iImage = m_pAsyncShellIcon->RetrieveIconIndex();
		m_pAsyncShellIcon->Release();
		m_pAsyncShellIcon = 0;
	}

	return m_iImage != II_NONE;
}

/*
 * Case insensitive sort helpers.
 */

inline char
ChCaseless( char ch )
{
	// lstrcmpi apparently uses lower case for caseless comparisons, so we
	// must also.
	return ( ch >= 'A' && ch <= 'Z' ) ? ( ch | 0x20 ) : ch;
}

enum CmpSortFlags
{
	CSF_NONE				= 0x00,
	CSF_WILDCARDS			= 0x01,
	CSF_CASEINSENSITIVE		= 0x02,
	CSF_STARSTOPSATSLASH	= 0x04,
};
inline CmpSortFlags operator|( CmpSortFlags const& a, CmpSortFlags const& b ) { return CmpSortFlags( int( a ) | int( b ) ); }
inline CmpSortFlags operator&( CmpSortFlags const& a, CmpSortFlags const& b ) { return CmpSortFlags( int( a ) & int( b ) ); }

static CmpSortFlags s_flagsDefault = CSF_CASEINSENSITIVE;

static int
cmpsort( char const* psz1, int _cch1, char const* psz2, int _cch2, CmpSortFlags flags )
{
	static unsigned char const c_rgCharsInOrder[256] =
	{
		0x00,	//	^@	//
		0x01,	//	^A	//
		0x02,	//	^B	//
		0x03,	//	^C	//
		0x04,	//	^D	//
		0x05,	//	^E	//
		0x06,	//	^F	//
		0x07,	//	^G	//
		0x08,	//	^H	//
		0x09,	//	^I	//
		0x0a,	//	^J	//
		0x0b,	//	^K	//
		0x0c,	//	^L	//
		0x0d,	//	^M	//
		0x0e,	//	^N	//
		0x0f,	//	^O	//
		0x10,	//	^P	//
		0x11,	//	^Q	//
		0x12,	//	^R	//
		0x13,	//	^S	//
		0x14,	//	^T	//
		0x15,	//	^U	//
		0x16,	//	^V	//
		0x17,	//	^W	//
		0x18,	//	^X	//
		0x19,	//	^Y	//
		0x1a,	//	^Z	//
		0x1b,	//	^[	//
		0x1c,	//	^\	//
		0x1d,	//	^]	//
		0x1e,	//	^^	//
		0x1f,	//	^_	//
		0x20,	//	SPC	//
		0x2e,	//	.	//
		0x2c,	//	,	//
		0x21,	//	!	//
		0x22,	//	"	//
		0x23,	//	#	//
		0x24,	//	$	//
		0x25,	//	%	//
		0x26,	//	&	//
		0x27,	//	'	//
		0x28,	//	(	//
		0x29,	//	)	//
		0x2a,	//	*	//
		0x2b,	//	+	//
		0x2f,	//	/	//
		0x3a,	//	:	//
		0x3b,	//	;	//
		0x3c,	//	<	//
		0x3d,	//	=	//
		0x3e,	//	>	//
		0x3f,	//	?	//
		0x5b,	//	[	//
		0x5c,	//	\	//
		0x5d,	//	]	//
		0x5e,	//	^	//
		0x60,	//	`	//
		0x40,	//	@	//
		0x7b,	//	{	//
		0x7c,	//	|	//
		0x7d,	//	}	//
		0x5f,	//	_	//
		0x7e,	//	~	//
		0x41,	//	A	//
		0x61,	//	a	//
		0x42,	//	B	//
		0x62,	//	b	//
		0x43,	//	C	//
		0x63,	//	c	//
		0x44,	//	D	//
		0x64,	//	d	//
		0x45,	//	E	//
		0x65,	//	e	//
		0x46,	//	F	//
		0x66,	//	f	//
		0x47,	//	G	//
		0x67,	//	g	//
		0x48,	//	H	//
		0x68,	//	h	//
		0x49,	//	I	//
		0x69,	//	i	//
		0x4a,	//	J	//
		0x6a,	//	j	//
		0x4b,	//	K	//
		0x6b,	//	k	//
		0x4c,	//	L	//
		0x6c,	//	l	//
		0x4d,	//	M	//
		0x6d,	//	m	//
		0x4e,	//	N	//
		0x6e,	//	n	//
		0x4f,	//	O	//
		0x6f,	//	o	//
		0x50,	//	P	//
		0x70,	//	p	//
		0x51,	//	Q	//
		0x71,	//	q	//
		0x52,	//	R	//
		0x72,	//	r	//
		0x53,	//	S	//
		0x73,	//	s	//
		0x54,	//	T	//
		0x74,	//	t	//
		0x55,	//	U	//
		0x75,	//	u	//
		0x56,	//	V	//
		0x76,	//	v	//
		0x57,	//	W	//
		0x77,	//	w	//
		0x58,	//	X	//
		0x78,	//	x	//
		0x59,	//	Y	//
		0x79,	//	y	//
		0x5a,	//	Z	//
		0x7a,	//	z	//
		0x30,	//	0	//
		0x31,	//	1	//
		0x32,	//	2	//
		0x33,	//	3	//
		0x34,	//	4	//
		0x35,	//	5	//
		0x36,	//	6	//
		0x37,	//	7	//
		0x38,	//	8	//
		0x39,	//	9	//
		0x2d,	//	-	//
		0x7f,	//		//
		0x80,	//		//
		0x81,	//		//
		0x82,	//		//
		0x83,	//		//
		0x84,	//		//
		0x85,	//		//
		0x86,	//		//
		0x87,	//		//
		0x88,	//		//
		0x89,	//		//
		0x8a,	//		//
		0x8b,	//		//
		0x8c,	//		//
		0x8d,	//		//
		0x8e,	//		//
		0x8f,	//		//
		0x90,	//		//
		0x91,	//		//
		0x92,	//		//
		0x93,	//		//
		0x94,	//		//
		0x95,	//		//
		0x96,	//		//
		0x97,	//		//
		0x98,	//		//
		0x99,	//		//
		0x9a,	//		//
		0x9b,	//		//
		0x9c,	//		//
		0x9d,	//		//
		0x9e,	//		//
		0x9f,	//		//
		0xa0,	//		//
		0xa1,	//		//
		0xa2,	//		//
		0xa3,	//		//
		0xa4,	//		//
		0xa5,	//		//
		0xa6,	//		//
		0xa7,	//		//
		0xa8,	//		//
		0xa9,	//		//
		0xaa,	//		//
		0xab,	//		//
		0xac,	//		//
		0xad,	//		//
		0xae,	//		//
		0xaf,	//		//
		0xb0,	//		//
		0xb1,	//		//
		0xb2,	//		//
		0xb3,	//		//
		0xb4,	//		//
		0xb5,	//		//
		0xb6,	//		//
		0xb7,	//		//
		0xb8,	//		//
		0xb9,	//		//
		0xba,	//		//
		0xbb,	//		//
		0xbc,	//		//
		0xbd,	//		//
		0xbe,	//		//
		0xbf,	//		//
		0xc0,	//		//
		0xc1,	//		//
		0xc2,	//		//
		0xc3,	//		//
		0xc4,	//		//
		0xc5,	//		//
		0xc6,	//		//
		0xc7,	//		//
		0xc8,	//		//
		0xc9,	//		//
		0xca,	//		//
		0xcb,	//		//
		0xcc,	//		//
		0xcd,	//		//
		0xce,	//		//
		0xcf,	//		//
		0xd0,	//		//
		0xd1,	//		//
		0xd2,	//		//
		0xd3,	//		//
		0xd4,	//		//
		0xd5,	//		//
		0xd6,	//		//
		0xd7,	//		//
		0xd8,	//		//
		0xd9,	//		//
		0xda,	//		//
		0xdb,	//		//
		0xdc,	//		//
		0xdd,	//		//
		0xde,	//		//
		0xdf,	//		//
		0xe0,	//		//
		0xe1,	//		//
		0xe2,	//		//
		0xe3,	//		//
		0xe4,	//		//
		0xe5,	//		//
		0xe6,	//		//
		0xe7,	//		//
		0xe8,	//		//
		0xe9,	//		//
		0xea,	//		//
		0xeb,	//		//
		0xec,	//		//
		0xed,	//		//
		0xee,	//		//
		0xef,	//		//
		0xf0,	//		//
		0xf1,	//		//
		0xf2,	//		//
		0xf3,	//		//
		0xf4,	//		//
		0xf5,	//		//
		0xf6,	//		//
		0xf7,	//		//
		0xf8,	//		//
		0xf9,	//		//
		0xfa,	//		//
		0xfb,	//		//
		0xfc,	//		//
		0xfd,	//		//
		0xfe,	//		//
		0xff,	//		//
	};

	static bool s_fOrder = false;
	static unsigned char s_rgOrder[sizeof( c_rgCharsInOrder )];
	CASSERT( sizeof( s_rgOrder ) == 256 );

	if( !s_fOrder )
	{
		Assert( c_rgCharsInOrder[0xff] == 0xff );
		Assert( c_rgCharsInOrder[0xfe] == 0xfe );
		for( unsigned ii = 0; ii < sizeof( c_rgCharsInOrder ); ii++ )
		{
			if( ii )
			{
				Assert2( s_rgOrder[c_rgCharsInOrder[ii]] != ii,
						 "BUG: duplicate char %u at index %u.",
						 c_rgCharsInOrder[ii], ii );
			}
			s_rgOrder[c_rgCharsInOrder[ii]] = ii;
		}
		s_fOrder = true;
	}

	//$ todo: (chrisant) Performance could be improved by just detecting null
	// terminators as we compare, but that would make it harder to support the
	// strnicmp usage pattern.
	if( _cch1 < 0 )
		_cch1 = lstrlen( psz1 ) + 1;
	if( _cch2 < 0 )
		_cch2 = lstrlen( psz2 ) + 1;

	int n = 0;

	Trace_( "[%s]/%d vs [%s]/%d == ", psz1, _cch1, psz2, _cch2 );

	if( _cch1 || _cch2 )
	{
		CASSERT( static_cast<unsigned char>( char( 0xff ) ) == 0xff );

		char const* const pszEnd1 = psz1 + _cch1;
		char const* const pszEnd2 = psz2 + _cch2;

		// First match fixed prefix.

		while( psz1 < pszEnd1 )
		{
			char const ch2Raw = psz2 < pszEnd2 ? *psz2 : 0;

			if( flags & CSF_WILDCARDS && *psz1 == '*' )
				break;

			if( flags & CSF_CASEINSENSITIVE )
			{
				char const ch1 = s_rgOrder[static_cast<unsigned char>( ChCaseless( *psz1 ) )];
				char const ch2 = s_rgOrder[static_cast<unsigned char>( ChCaseless( ch2Raw ) )];
				n = ( static_cast<int>( static_cast<unsigned char>( ch1 ) ) -
					  static_cast<int>( static_cast<unsigned char>( ch2 ) ) );
			}
			else
			{
				n = ( static_cast<int>( static_cast<unsigned char>( *psz1 ) ) -
					  static_cast<int>( static_cast<unsigned char>( *psz2 ) ) );
			}

			if( flags & CSF_WILDCARDS && *psz1 == '?' && ch2Raw )
				n = 0;

			psz1++;
			if( psz2 < pszEnd2 )
				psz2++;

			if( n )
				break;
		}

		// Now handle star wildcards.

		if( flags & CSF_WILDCARDS && !n && psz1 < pszEnd1 && *psz1 == '*' )
		{
			struct Param
			{
				size_t		iStart;
				size_t		iEnd;
			};
			struct Params
			{
				char const*	buffer;
				Param		rgParam[5];
			};
			struct BackupStruct
			{
				char const*	pch;
				Param*		pParam;
			};

			Params params;
			params.buffer = psz2;

			BackupStruct rgBackups[5];
			BackupStruct* pNextBackup = rgBackups;
			BackupStruct* pBackup;

			for( ;; )
			{
				pBackup = pNextBackup;

				// Check for end of wildcard expression string.

				if( psz1 >= pszEnd1 )
				{
					Trace( "0  (Has wildcards)" );
					return 0;

LRetry:
					for( ;; --pBackup )
					{
						if( pBackup <= rgBackups )
						{
							Trace( "-1  (Failed to match, has wildcards)" );
							return -1;
						}

						psz1 = pBackup[-1].pch;
						psz2 = params.buffer + --( pBackup[-1].pParam->iEnd );

						Assert( psz1 < pszEnd1 );

						if( *psz1 != '*' && *psz1 != '?' )
						{
							Assert( psz2 < pszEnd2 );

							// Optimization:  If *psz1 != *psz2 we'll get here
							// again through a circuitous path just to
							// decrement pBackup[-1].pParam->iEnd.  So the
							// loop below backtracks more efficiently.

							char const* const pchStart = params.buffer + pBackup[-1].pParam->iStart;
							if( flags & CSF_CASEINSENSITIVE )
							{
								while( psz2 >= pchStart )
								{
									char const ch1 = s_rgOrder[static_cast<unsigned char>( ChCaseless( *psz1 ) )];
									char const ch2 = s_rgOrder[static_cast<unsigned char>( ChCaseless( *psz2 ) )];
									if( ch1 == ch2 )
										break;
									psz2--;
								}
							}
							else
							{
								while( psz2 >= pchStart )
								{
									if( *psz1 == *psz2 )
										break;
									psz2--;
								}
							}
							pBackup[-1].pParam->iEnd = psz2 - params.buffer;
						}

						if( psz2 >= params.buffer + pBackup[-1].pParam->iStart )
						{
							pNextBackup = pBackup;
							break;
						}
					}
					continue;
				}

				// Match characters.

				if( *psz1 == '*' )
				{
					pNextBackup++;
					if( pNextBackup - rgBackups > DimensionOf( rgBackups ) )
					{
						// Too many wildcards; give up and fail.
						Trace( "1  (Too many wildcards!)" );
						return 1;
					}

					pBackup->pParam = params.rgParam + ( pBackup - rgBackups );
					pBackup->pParam->iStart = psz2 - params.buffer;

					while( psz2 < pszEnd2 && ( !( flags & CSF_STARSTOPSATSLASH ) || *psz2 != '\\' ) )
						psz2++;

					while( psz1 < pszEnd1 && *psz1 == '*' )
						psz1++;

					pBackup->pParam->iEnd = psz2 - params.buffer;
					pBackup->pch = psz1;
				}
				else
				{
					while( psz1 < pszEnd1 )
					{
						if( *psz1 == '*' )
							break;

						if( psz2 >= pszEnd2 )
							goto LRetry;

						if( flags & CSF_WILDCARDS )
						{
							char const ch1 = s_rgOrder[static_cast<unsigned char>( ChCaseless( *psz1 ) )];
							char const ch2 = s_rgOrder[static_cast<unsigned char>( ChCaseless( *psz2 ) )];
							n = ( static_cast<int>( static_cast<unsigned char>( ch1 ) ) -
								  static_cast<int>( static_cast<unsigned char>( ch2 ) ) );
						}
						else
						{
							n = ( static_cast<int>( static_cast<unsigned char>( *psz1 ) ) -
								  static_cast<int>( static_cast<unsigned char>( *psz2 ) ) );
						}

						if( flags & CSF_WILDCARDS && *psz1 == '?' && *psz2 )
							n = 0;

						psz1++;
						psz2++;

						if( n )
							goto LRetry;
					}
				}
			}
		}
	}

	Trace( "%d", n );
	return n;
}

static int
Cmp_NoCase( char const* psz1, char const* psz2 )
{
	return cmpsort( psz1, -1, psz2, -1, s_flagsDefault );
}

static int
Cmp_WildNoCase( char const* psz1, char const* psz2 )
{
	return cmpsort( psz1, -1, psz2, -1, s_flagsDefault|CSF_WILDCARDS );
}

static int
Cmp_Lookup( char const* pszPartialFile, FileInfo const* pfi )
{
	char const* const psz1 = pszPartialFile;
	char const* const psz2 = pfi->m_sFile.Text();
	// Note that the null terminators are excluded from the comparison.
	return cmpsort( psz1, lstrlen( psz1 ), psz2, lstrlen( psz2 ), s_flagsDefault|CSF_WILDCARDS );
}

static int
Cmp_NonWildPrefix( char const* pszPartialFile, FileInfo const* pfi )
{
	char const* const psz1 = pszPartialFile;
	char const* const psz2 = pfi->m_sFile.Text();
	char const* const pszWild = strpbrk( pszPartialFile, "*?" );
	Assert( pszWild );
	// Note that the null terminators are excluded from the comparison.
	return cmpsort( psz1, pszWild - psz1, psz2, lstrlen( psz2 ), s_flagsDefault|CSF_WILDCARDS );
}

static int
Cmp_SortByName( void const* pv1, void const* pv2 )
{
	FileInfo const* const pfi1 = static_cast<FileInfo const*>( pv1 );
	FileInfo const* const pfi2 = static_cast<FileInfo const*>( pv2 );
	int n;
	n = int( pfi2->m_attr & FILE_ATTRIBUTE_DIRECTORY ) - int( pfi1->m_attr & FILE_ATTRIBUTE_DIRECTORY );
	if( n )
		return n;
	n = Cmp_NoCase( pfi1->m_sFile.Text(), pfi2->m_sFile.Text() );
	if( n )
		return n;
	n = Cmp_NoCase( pfi1->m_sPath.Text(), pfi2->m_sPath.Text() );
	if( n )
		return n;
	return 0;
}

static int
Cmp_SortByPath( void const* pv1, void const* pv2 )
{
	FileInfo const* const pfi1 = static_cast<FileInfo const*>( pv1 );
	FileInfo const* const pfi2 = static_cast<FileInfo const*>( pv2 );
	int n;
	n = int( pfi2->m_attr & FILE_ATTRIBUTE_DIRECTORY ) - int( pfi1->m_attr & FILE_ATTRIBUTE_DIRECTORY );
	if( n )
		return n;
	n = Cmp_NoCase( pfi1->m_sPath.Text(), pfi2->m_sPath.Text() );
	if( n )
		return n;
	n = Cmp_NoCase( pfi1->m_sFile.Text(), pfi2->m_sFile.Text() );
	if( n )
		return n;
	return 0;
}

/*
 * Helper functions.
 */

char const*
GetFileName( char const* pszFile )
{
	char const* pszName = pszFile;

	for( char const* pszWalk = pszFile; *pszWalk; pszWalk++ )
		if( '/' == *pszWalk || '\\' == *pszWalk )
			pszName = pszWalk + 1;

	return pszName;
}

char const*
GetExtension( char const* pszFile )
{
	char const* pszDot;

	for( pszDot = 0; *pszFile; pszFile++ )
		switch( *pszFile )
		{
		case '.':
			pszDot = pszFile;
			break;
		case ' ':
		case ':':
		case '\\':
		case '/':
			pszDot = 0;
			break;
		}

	return pszDot;
}

/*
 * File list public functions.
 */

void VSAPI
FlatFile_Uninitialize()
{
	delete s_prgMatches;
	s_prgMatches = 0;

	for( unsigned ii = DimensionOf( s_rgprgFiles ); ii--; )
	{
		delete s_rgprgFiles[ii];
		s_rgprgFiles[ii] = 0;
	}
}

void VSAPI
FlatFile_SetOptions( int sort_by_path,
					 int case_sensitive,
					 int always_substring_match )
{
	s_fSortByPath = !!sort_by_path;
	s_flagsDefault = ( !case_sensitive ? CSF_CASEINSENSITIVE : CSF_NONE );
	s_fAlwaysSubstringMatch = !!always_substring_match;
}

void VSAPI
FlatFile_SetMode( int iMode )
{
	Assert( iMode >= 0 );
	Assert( iMode < 3 );

	delete s_prgMatches;
	s_prgMatches = 0;

	s_iMode = iMode;
}

int VSAPI
FlatFile_GetMode()
{
	return s_iMode;
}

void VSAPI
FlatFile_ClearFiles()
{
	delete s_prgMatches;
	s_prgMatches = 0;

	delete s_rgprgFiles[s_iMode];
	s_rgprgFiles[s_iMode] = 0;
}

int VSAPI
FlatFile_AllocateFileList( unsigned cFiles )
{
	FlatFile_ClearFiles();

	s_rgprgFiles[s_iMode] = new Rg<FileInfo>;
	if( !s_rgprgFiles[s_iMode] )
		return -1;
	if( !s_rgprgFiles[s_iMode]->EnsureAllocation( cFiles ) )
		return -1;
	return 0;
}

int VSAPI
FlatFile_AppendFile( VSHVAR hvarFilename )
{
	if( !s_rgprgFiles[s_iMode] )
		return -1;

	char const* const pszPath = vsHvarGetZ( hvarFilename );
	if( !pszPath )
		return -1;

	FileInfo* p;
	if( !s_rgprgFiles[s_iMode]->Append( &p ) )
		return -1;

	char const* const pszFile = GetFileName( pszPath );
	size_t const cchPath = ( pszFile - pszPath );

	p->m_sFile.Set( pszFile );
	p->m_sPath.Set( pszPath, cchPath );
	p->m_attr = 0;
#if 0
	ZeroMemory( &p->m_ft, sizeof( p->m_ft ) );
	ZeroMemory( &p->m_uliFileSize, sizeof( p->m_uliFileSize ) );
#endif

	return 0;
}

void VSAPI
FlatFile_SortFiles()
{
	if( s_rgprgFiles[s_iMode] )
	{
		s_rgprgFiles[s_iMode]->Sort( s_fSortByPath ? Cmp_SortByPath : Cmp_SortByName );
		Trace( "FlatFile_SortFiles:  sorted %u files.", s_rgprgFiles[s_iMode]->Count() );
	}
}

/*
 * File list internal functions.
 */

unsigned
FlatFile_Count()
{
	if( s_prgMatches )
		return s_prgMatches->Count();
	if( s_rgprgFiles[s_iMode] )
		return s_rgprgFiles[s_iMode]->Count();
	return 0;
}

int
FlatFile_AppendFileData( char const* pszDir, WIN32_FIND_DATA const* pfd )
{
	Assert( s_iMode == 0 );

	if( !s_rgprgFiles[s_iMode] )
		return -1;

	FileInfo* p;
	if( !s_rgprgFiles[s_iMode]->Append( &p ) )
		return -1;

	Assert( *pszDir && *pfd->cFileName );
	p->m_sFile.Set( pfd->cFileName );
	p->m_sPath.Set( pszDir );
	p->m_attr = pfd->dwFileAttributes;
	p->m_ft = pfd->ftLastWriteTime;
	p->m_uliFileSize.LowPart = pfd->nFileSizeLow;
	p->m_uliFileSize.HighPart = pfd->nFileSizeHigh;

	if( pfd->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
		p->m_sFile.Append( "\\" );

	return 0;
}

FileInfo*
FlatFile_GetFile( unsigned index )
{
	if( s_prgMatches )
	{
		if( index >= s_prgMatches->Count() )
			return 0;
		return *s_prgMatches->Item( index );
	}

	if( s_rgprgFiles[s_iMode] )
	{
		if( index >= s_rgprgFiles[s_iMode]->Count() )
			return 0;
		return s_rgprgFiles[s_iMode]->Item( index );
	}

	return 0;
}

/*
 * Match filter internal functions.
 */

void
FlatFile_Filter( char const* pszFilter )
{
	// Reset the matches array.

	delete s_prgMatches;
	s_prgMatches = 0;

	// Prepare to scan the appropriate files list array.

	Rg<FileInfo>* const prg = s_rgprgFiles[s_iMode];

	if( !prg || !pszFilter || !*pszFilter )
		return;

	// At this point we are filtering for sure, so go ahead and create a new
	// matches array.

	s_prgMatches = new Rg<FileInfo*>;

	// Split the lookup text into path and file components because the list is
	// sorted by file rather than by path.

	Str sFile;
	Str sPath;
	char const* const _pszFile = GetFileName( pszFilter );
	sFile.Set( _pszFile );
	if( _pszFile > pszFilter )
	{
		sPath.Set( "*" );
		sPath.Append( pszFilter, _pszFile - pszFilter );
	}
	if( s_fAlwaysSubstringMatch )
	{
		sFile.Set( "*" );
		sFile.Append( _pszFile );
		sFile.Append( "*" );
		if( sPath.Length() )
		{
			Assert( sPath.Length() > 0 );
			Assert( sPath.Text()[sPath.Length() - 1] == '\\' );
			sPath.SetLength( sPath.Length() - 1 );
			sPath.Append( "*\\" );
		}
	}

	char const* const pszFile = sFile.Text();

	// Try to efficiently skip to the first possible match.

	unsigned iStartAt = 0;
	bool const fBeginsWithWildcard = ( *pszFile == '*' || *pszFile == '?' );
	bool const fHasWildcards = ( strchr( pszFile, '*' ) || strchr( pszFile, '?' ) );

	if( *pszFile &&						// No point skipping if no file component.
		!fBeginsWithWildcard &&			// Can't skip if first character is a wildcard.
		!s_fSortByPath &&				// Can't skip when sorted by path.
		s_iMode != 0 &&					// "Any" mode is essentially unsorted (dirs, then files).
		!prg->FSearch( reinterpret_cast<CompareFn>( fHasWildcards ? Cmp_NonWildPrefix : Cmp_Lookup ), pszFile, &iStartAt ) )
		return;

	// Iterate looking for matches.

	FileInfo** pp;

	while( iStartAt < prg->Count() )
	{
		FileInfo const* const p = prg->Item( iStartAt++ );

		// Test if the item matches the lookup text.

		bool const fMatchPath = ( !sPath.Length() || 0 == Cmp_WildNoCase( sPath.Text(), p->m_sPath.Text() ) );
		bool const fMatchFile = ( fMatchPath && ( !*pszFile || 0 == Cmp_Lookup( pszFile, p ) ) );

		// If the file didn't match then some processing is needed to figure
		// out when to exit the loop.

		if( !fMatchFile )
		{
			// If the lookup text has a path, then we have to scan the entire
			// array because it is sorted by file not by path.
			if( sPath.Length() )
				continue;
			// If the lookup text has wildcards but does not begin with a
			// wildcard then we have to scan the entire array because it is
			// sorted by file not by path.
			if( fHasWildcards && !fBeginsWithWildcard )
				continue;
			// When sorted by path we have to scan the entire array.
			if( s_fSortByPath )
				continue;
			// If the lookup text has wildcards and begins with a non-wildcard
			// character then we only have to scan while the prefix matches.
			if( fHasWildcards && fBeginsWithWildcard )
				if( 0 == Cmp_NonWildPrefix( pszFile, p ) )
					continue;
			// Otherwise failure to match means we're done.
			break;
		}

		// Append match to the filtered array.

		s_prgMatches->Append( &pp );
		if( !pp )
			break;
		*pp = const_cast<FileInfo*>( p );
	}
}

bool
FlatFile_IsFiltered()
{
	return !!s_prgMatches;
}

