/*
 * Utilities.
 */

#pragma once

#include "std.h"

/*
 * Resource helpers.
 */

HINSTANCE ResourceModule();

inline HMENU LoadMenuU( UINT idr ) { return LoadMenu( ResourceModule(), MAKEINTRESOURCE( idr ) ); }
inline INT_PTR DialogBoxParamU( UINT idDlg, HWND hwndParent, DLGPROC dlgproc, LPARAM initParam ) { return DialogBoxParam( ResourceModule(), MAKEINTRESOURCE( idDlg ), hwndParent, dlgproc, initParam ); }
inline HWND CreateDialogParamU( UINT idDlg, HWND hwndParent, DLGPROC dlgproc, LPARAM initParam ) { return CreateDialogParam( ResourceModule(), MAKEINTRESOURCE( idDlg ), hwndParent, dlgproc, initParam ); }
inline HACCEL LoadAcceleratorsU( UINT idrAccel ) { return LoadAccelerators( ResourceModule(), MAKEINTRESOURCE( idrAccel ) ); }

/*
 * Theme helpers.
 */

bool ThemeSubclassEdit( HWND hwnd );
bool ThemeExplorerTheme( HWND hwnd );

/*
 * Layout helpers.
 */

class DeferPosXact
{
public:
					DeferPosXact() : m_hdwp( 0 ) {}
					~DeferPosXact() { End(); }

	bool			Begin( int cWindows ) { Assert( !m_hdwp ); m_hdwp = BeginDeferWindowPos( cWindows ); return !!m_hdwp; }
	void			End() { if( m_hdwp ) EndDeferWindowPos( m_hdwp ); }

	bool			DeferWindowPos( HWND hwnd, LPCRECT prc, UINT grf = SWP_NOZORDER|SWP_DRAWFRAME|SWP_NOACTIVATE|SWP_NOCOPYBITS, HWND hwndAfter = 0 );
	bool			DeferWindowPos( HWND hwnd, int x, int y, int cx, int cy, UINT grf = SWP_NOZORDER|SWP_DRAWFRAME|SWP_NOACTIVATE|SWP_NOCOPYBITS, HWND hwndAfter = 0 );
	bool			DeferWindowPos( HWND hwnd, POINT pt );
	bool			DeferWindowPos( HWND hwnd, HWND hwndInsertAfter, int x, int y, int cx, int cy, UINT ulFlags );


	void			AdjustSize( HWND hwnd, int dx, int dy );
	void			AdjustPos( HWND hwnd, int dx, int dy );

private:
	HDWP			m_hdwp;
};

inline bool
DeferPosXact::DeferWindowPos( HWND hwnd, LPCRECT prc, UINT grf, HWND hwndAfter )
{
	return DeferWindowPos( hwnd, hwndAfter, prc->left, prc->top, prc->right - prc->left, prc->bottom - prc->top, grf );
}

inline bool
DeferPosXact::DeferWindowPos( HWND hwnd, int x, int y, int cx, int cy, UINT grf, HWND hwndAfter )
{
	return DeferWindowPos( hwnd, hwndAfter, x, y, cx, cy, grf );
}

inline bool
DeferPosXact::DeferWindowPos( HWND hwnd, POINT pt )
{
	return DeferWindowPos( hwnd, 0, pt.x, pt.y, 0, 0, SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOCOPYBITS );
}

/*
 * Color blending.
 */

int Luminance( COLORREF cr );
COLORREF BlendColors( COLORREF cr1, COLORREF cr2, BYTE alpha );
COLORREF MakeBlendedGray( HDC hdc );
COLORREF MakeBlendedGray( HDC hdc, BYTE alpha, bool fFavorLight );

